#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


__version__ = '5.0.0'
__author__ = 'Marco Caspers'
__author_email__ = 'samadevteam@westcon.com'
__copyright__ = 'Copyright 2023 Westcon-Comstor'
__license__ = 'MIT License'